@@ -6,10 +6,10 @@
 from __future__ import annotations
 
 import collections
+import dataclasses
+from collections.abc import Container, Iterable
+from typing import TYPE_CHECKING
 
-from typing import Callable, Dict, Iterable, List, Optional, Tuple, TYPE_CHECKING
-
-from coverage.debug import AutoReprMixin
 from coverage.exceptions import ConfigError
 from coverage.misc import nice_pair
 from coverage.types import TArc, TLineNo
@@ -19,45 +19,93 @@ if TYPE_CHECKING:
     from coverage.plugin import FileReporter
 
 
+def analysis_from_file_reporter(
+    data: CoverageData,
+    precision: int,
+    file_reporter: FileReporter,
+    filename: str,
+) -> Analysis:
+    """Create an Analysis from a FileReporter."""
+    has_arcs = data.has_arcs()
+    statements = file_reporter.lines()
+    excluded = file_reporter.excluded_lines()
+    executed = file_reporter.translate_lines(data.lines(filename) or [])
+
+    if has_arcs:
+        arc_possibilities_set = file_reporter.arcs()
+        arcs: Iterable[TArc] = data.arcs(filename) or []
+        arcs = file_reporter.translate_arcs(arcs)
+
+        # Reduce the set of arcs to the ones that could be branches.
+        dests = collections.defaultdict(set)
+        for fromno, tono in arc_possibilities_set:
+            dests[fromno].add(tono)
+        single_dests = {
+            fromno: list(tonos)[0]
+            for fromno, tonos in dests.items()
+            if len(tonos) == 1
+        }
+        new_arcs = set()
+        for fromno, tono in arcs:
+            if fromno != tono:
+                new_arcs.add((fromno, tono))
+            else:
+                if fromno in single_dests:
+                    new_arcs.add((fromno, single_dests[fromno]))
+
+        arcs_executed_set = file_reporter.translate_arcs(new_arcs)
+        exit_counts = file_reporter.exit_counts()
+        no_branch = file_reporter.no_branch_lines()
+    else:
+        arc_possibilities_set = set()
+        arcs_executed_set = set()
+        exit_counts = {}
+        no_branch = set()
+
+    return Analysis(
+        precision=precision,
+        filename=filename,
+        has_arcs=has_arcs,
+        statements=statements,
+        excluded=excluded,
+        executed=executed,
+        arc_possibilities_set=arc_possibilities_set,
+        arcs_executed_set=arcs_executed_set,
+        exit_counts=exit_counts,
+        no_branch=no_branch,
+    )
+
+
+@dataclasses.dataclass
 class Analysis:
     """The results of analyzing a FileReporter."""
 
-    def __init__(
-        self,
-        data: CoverageData,
-        precision: int,
-        file_reporter: FileReporter,
-        file_mapper: Callable[[str], str],
-    ) -> None:
-        self.data = data
-        self.file_reporter = file_reporter
-        self.filename = file_mapper(self.file_reporter.filename)
-        self.statements = self.file_reporter.lines()
-        self.excluded = self.file_reporter.excluded_lines()
+    precision: int
+    filename: str
+    has_arcs: bool
+    statements: set[TLineNo]
+    excluded: set[TLineNo]
+    executed: set[TLineNo]
+    arc_possibilities_set: set[TArc]
+    arcs_executed_set: set[TArc]
+    exit_counts: dict[TLineNo, int]
+    no_branch: set[TLineNo]
 
-        # Identify missing statements.
-        executed: Iterable[TLineNo]
-        executed = self.data.lines(self.filename) or []
-        executed = self.file_reporter.translate_lines(executed)
-        self.executed = executed
+    def __post_init__(self) -> None:
+        self.arc_possibilities = sorted(self.arc_possibilities_set)
+        self.arcs_executed = sorted(self.arcs_executed_set)
         self.missing = self.statements - self.executed
 
-        if self.data.has_arcs():
-            self._arc_possibilities = sorted(self.file_reporter.arcs())
-            self.exit_counts = self.file_reporter.exit_counts()
-            self.no_branch = self.file_reporter.no_branch_lines()
+        if self.has_arcs:
             n_branches = self._total_branches()
             mba = self.missing_branch_arcs()
             n_partial_branches = sum(len(v) for k,v in mba.items() if k not in self.missing)
             n_missing_branches = sum(len(v) for k,v in mba.items())
         else:
-            self._arc_possibilities = []
-            self.exit_counts = {}
-            self.no_branch = set()
             n_branches = n_partial_branches = n_missing_branches = 0
 
         self.numbers = Numbers(
-            precision=precision,
+            precision=self.precision,
             n_files=1,
             n_statements=len(self.statements),
             n_excluded=len(self.excluded),
@@ -67,6 +115,50 @@ class Analysis:
             n_missing_branches=n_missing_branches,
         )
 
+    def narrow(self, lines: Container[TLineNo]) -> Analysis:
+        """Create a narrowed Analysis.
+
+        The current analysis is copied to make a new one that only considers
+        the lines in `lines`.
+        """
+
+        statements = {lno for lno in self.statements if lno in lines}
+        excluded = {lno for lno in self.excluded if lno in lines}
+        executed = {lno for lno in self.executed if lno in lines}
+
+        if self.has_arcs:
+            arc_possibilities_set = {
+                (a, b) for a, b in self.arc_possibilities_set
+                if a in lines or b in lines
+            }
+            arcs_executed_set = {
+                (a, b) for a, b in self.arcs_executed_set
+                if a in lines or b in lines
+            }
+            exit_counts = {
+                lno: num for lno, num in self.exit_counts.items()
+                if lno in lines
+            }
+            no_branch = {lno for lno in self.no_branch if lno in lines}
+        else:
+            arc_possibilities_set = set()
+            arcs_executed_set = set()
+            exit_counts = {}
+            no_branch = set()
+
+        return Analysis(
+            precision=self.precision,
+            filename=self.filename,
+            has_arcs=self.has_arcs,
+            statements=statements,
+            excluded=excluded,
+            executed=executed,
+            arc_possibilities_set=arc_possibilities_set,
+            arcs_executed_set=arcs_executed_set,
+            exit_counts=exit_counts,
+            no_branch=no_branch,
+        )
+
     def missing_formatted(self, branches: bool = False) -> str:
         """The missing line numbers, formatted nicely.
 
@@ -75,58 +167,24 @@ class Analysis:
         If `branches` is true, includes the missing branch arcs also.
 
         """
-        if branches and self.has_arcs():
+        if branches and self.has_arcs:
             arcs = self.missing_branch_arcs().items()
         else:
             arcs = None
 
         return format_lines(self.statements, self.missing, arcs=arcs)
 
-    def has_arcs(self) -> bool:
-        """Were arcs measured in this result?"""
-        return self.data.has_arcs()
-
-    def arc_possibilities(self) -> List[TArc]:
-        """Returns a sorted list of the arcs in the code."""
-        return self._arc_possibilities
-
-    def arcs_executed(self) -> List[TArc]:
-        """Returns a sorted list of the arcs actually executed in the code."""
-        executed: Iterable[TArc]
-        executed = self.data.arcs(self.filename) or []
-        executed = self.file_reporter.translate_arcs(executed)
-        return sorted(executed)
-
-    def arcs_missing(self) -> List[TArc]:
+    def arcs_missing(self) -> list[TArc]:
         """Returns a sorted list of the un-executed arcs in the code."""
-        possible = self.arc_possibilities()
-        executed = self.arcs_executed()
         missing = (
-            p for p in possible
-                if p not in executed
+            p for p in self.arc_possibilities
+                if p not in self.arcs_executed_set
                     and p[0] not in self.no_branch
                     and p[1] not in self.excluded
         )
         return sorted(missing)
 
-    def arcs_unpredicted(self) -> List[TArc]:
-        """Returns a sorted list of the executed arcs missing from the code."""
-        possible = self.arc_possibilities()
-        executed = self.arcs_executed()
-        # Exclude arcs here which connect a line to itself.  They can occur
-        # in executed data in some cases.  This is where they can cause
-        # trouble, and here is where it's the least burden to remove them.
-        # Also, generators can somehow cause arcs from "enter" to "exit", so
-        # make sure we have at least one positive value.
-        unpredicted = (
-            e for e in executed
-                if e not in possible
-                    and e[0] != e[1]
-                    and (e[0] > 0 or e[1] > 0)
-        )
-        return sorted(unpredicted)
-
-    def _branch_lines(self) -> List[TLineNo]:
+    def _branch_lines(self) -> list[TLineNo]:
         """Returns a list of line numbers that have more than one exit."""
         return [l1 for l1,count in self.exit_counts.items() if count > 1]
 
@@ -134,7 +192,7 @@ class Analysis:
         """How many total branches are there?"""
         return sum(count for count in self.exit_counts.values() if count > 1)
 
-    def missing_branch_arcs(self) -> Dict[TLineNo, List[TLineNo]]:
+    def missing_branch_arcs(self) -> dict[TLineNo, list[TLineNo]]:
         """Return arcs that weren't executed from branch lines.
 
         Returns {l1:[l2a,l2b,...], ...}
@@ -144,29 +202,35 @@ class Analysis:
         branch_lines = set(self._branch_lines())
         mba = collections.defaultdict(list)
         for l1, l2 in missing:
+            assert l1 != l2, f"In {self.filename}, didn't expect {l1} == {l2}"
             if l1 in branch_lines:
                 mba[l1].append(l2)
         return mba
 
-    def executed_branch_arcs(self) -> Dict[TLineNo, List[TLineNo]]:
+    def executed_branch_arcs(self) -> dict[TLineNo, list[TLineNo]]:
         """Return arcs that were executed from branch lines.
 
+        Only include ones that we considered possible.
+
         Returns {l1:[l2a,l2b,...], ...}
 
         """
-        executed = self.arcs_executed()
         branch_lines = set(self._branch_lines())
         eba = collections.defaultdict(list)
-        for l1, l2 in executed:
+        for l1, l2 in self.arcs_executed:
+            assert l1 != l2, f"Oops: Didn't think this could happen: {l1 = }, {l2 = }"
+            if (l1, l2) not in self.arc_possibilities_set:
+                continue
             if l1 in branch_lines:
                 eba[l1].append(l2)
         return eba
 
-    def branch_stats(self) -> Dict[TLineNo, Tuple[int, int]]:
+    def branch_stats(self) -> dict[TLineNo, tuple[int, int]]:
         """Get stats about branches.
 
         Returns a dict mapping line numbers to a tuple:
         (total_exits, taken_exits).
+
         """
 
         missing_arcs = self.missing_branch_arcs()
@@ -178,7 +242,8 @@ class Analysis:
         return stats
 
 
-class Numbers(AutoReprMixin):
+@dataclasses.dataclass
+class Numbers:
     """The numerical results of measuring coverage.
 
     This holds the basic statistics from `Analysis`, and is used to roll
@@ -186,36 +251,14 @@ class Numbers(AutoReprMixin):
 
     """
 
-    def __init__(
-        self,
-        precision: int = 0,
-        n_files: int = 0,
-        n_statements: int = 0,
-        n_excluded: int = 0,
-        n_missing: int = 0,
-        n_branches: int = 0,
-        n_partial_branches: int = 0,
-        n_missing_branches: int = 0,
-    ) -> None:
-        assert 0 <= precision < 10
-        self._precision = precision
-        self._near0 = 1.0 / 10**precision
-        self._near100 = 100.0 - self._near0
-        self.n_files = n_files
-        self.n_statements = n_statements
-        self.n_excluded = n_excluded
-        self.n_missing = n_missing
-        self.n_branches = n_branches
-        self.n_partial_branches = n_partial_branches
-        self.n_missing_branches = n_missing_branches
-
-    def init_args(self) -> List[int]:
-        """Return a list for __init__(*args) to recreate this object."""
-        return [
-            self._precision,
-            self.n_files, self.n_statements, self.n_excluded, self.n_missing,
-            self.n_branches, self.n_partial_branches, self.n_missing_branches,
-        ]
+    precision: int = 0
+    n_files: int = 0
+    n_statements: int = 0
+    n_excluded: int = 0
+    n_missing: int = 0
+    n_branches: int = 0
+    n_partial_branches: int = 0
+    n_missing_branches: int = 0
 
     @property
     def n_executed(self) -> int:
@@ -246,52 +289,26 @@ class Numbers(AutoReprMixin):
         result in either "0" or "100".
 
         """
-        return self.display_covered(self.pc_covered)
-
-    def display_covered(self, pc: float) -> str:
-        """Return a displayable total percentage, as a string.
-
-        Note that "0" is only returned when the value is truly zero, and "100"
-        is only returned when the value is truly 100.  Rounding can never
-        result in either "0" or "100".
-
-        """
-        if 0 < pc < self._near0:
-            pc = self._near0
-        elif self._near100 < pc < 100:
-            pc = self._near100
-        else:
-            pc = round(pc, self._precision)
-        return "%.*f" % (self._precision, pc)
-
-    def pc_str_width(self) -> int:
-        """How many characters wide can pc_covered_str be?"""
-        width = 3   # "100"
-        if self._precision > 0:
-            width += 1 + self._precision
-        return width
+        return display_covered(self.pc_covered, self.precision)
 
     @property
-    def ratio_covered(self) -> Tuple[int, int]:
+    def ratio_covered(self) -> tuple[int, int]:
         """Return a numerator and denominator for the coverage ratio."""
         numerator = self.n_executed + self.n_executed_branches
         denominator = self.n_statements + self.n_branches
         return numerator, denominator
 
     def __add__(self, other: Numbers) -> Numbers:
-        nums = Numbers(precision=self._precision)
-        nums.n_files = self.n_files + other.n_files
-        nums.n_statements = self.n_statements + other.n_statements
-        nums.n_excluded = self.n_excluded + other.n_excluded
-        nums.n_missing = self.n_missing + other.n_missing
-        nums.n_branches = self.n_branches + other.n_branches
-        nums.n_partial_branches = (
-            self.n_partial_branches + other.n_partial_branches
+        return Numbers(
+            self.precision,
+            self.n_files + other.n_files,
+            self.n_statements + other.n_statements,
+            self.n_excluded + other.n_excluded,
+            self.n_missing + other.n_missing,
+            self.n_branches + other.n_branches,
+            self.n_partial_branches + other.n_partial_branches,
+            self.n_missing_branches + other.n_missing_branches,
         )
-        nums.n_missing_branches = (
-            self.n_missing_branches + other.n_missing_branches
-        )
-        return nums
 
     def __radd__(self, other: int) -> Numbers:
         # Implementing 0+Numbers allows us to sum() a list of Numbers.
@@ -299,16 +316,34 @@ class Numbers(AutoReprMixin):
         return self
 
 
+def display_covered(pc: float, precision: int) -> str:
+    """Return a displayable total percentage, as a string.
+
+    Note that "0" is only returned when the value is truly zero, and "100"
+    is only returned when the value is truly 100.  Rounding can never
+    result in either "0" or "100".
+
+    """
+    near0 = 1.0 / 10 ** precision
+    if 0 < pc < near0:
+        pc = near0
+    elif (100.0 - near0) < pc < 100:
+        pc = 100.0 - near0
+    else:
+        pc = round(pc, precision)
+    return "%.*f" % (precision, pc)
+
+
 def _line_ranges(
     statements: Iterable[TLineNo],
     lines: Iterable[TLineNo],
-) -> List[Tuple[TLineNo, TLineNo]]:
+) -> list[tuple[TLineNo, TLineNo]]:
     """Produce a list of ranges for `format_lines`."""
     statements = sorted(statements)
     lines = sorted(lines)
 
     pairs = []
-    start = None
+    start: TLineNo | None = None
     lidx = 0
     for stmt in statements:
         if lidx >= len(lines):
@@ -329,7 +364,7 @@ def _line_ranges(
 def format_lines(
     statements: Iterable[TLineNo],
     lines: Iterable[TLineNo],
-    arcs: Optional[Iterable[Tuple[TLineNo, List[TLineNo]]]] = None,
+    arcs: Iterable[tuple[TLineNo, list[TLineNo]]] | None = None,
 ) -> str:
     """Nicely format a list of line numbers.
 
@@ -357,7 +392,7 @@ def format_lines(
                     dest = (ex if ex > 0 else "exit")
                     line_items.append((line, f"{line}->{dest}"))
 
-    ret = ', '.join(t[-1] for t in sorted(line_items))
+    ret = ", ".join(t[-1] for t in sorted(line_items))
     return ret
 
 
